#pragma once

// Genode includes
#include <base/heap.h>
#include <libc/component.h>
#include <util/list.h>

// gapfruit inculdes
#include <util/hexdump.h>

// local includes
#include "command.h"
#include "exception.h"
#include "payload.h"
#include "socket_wrapper.h"


namespace Hsm {
	class Message;
}

class Hsm::Message
{
	private:
		Command _command = Command::CryptoCommand;
		uint32_t _status = 0;
		Genode::Fifo<Payload> _payloads { };
		Genode::Heap& _heap;
		uint8_t* _data_buffer = nullptr;
		uint32_t _data_length = 0;

	public:
		Command command() const { return _command; }
		uint32_t status() const { return _status; }

		Message(const Message&) = delete;
		Message operator=(const Message&) = delete;

		Message(Genode::Heap& heap, const Command command)
			: _command(command), _heap(heap)
		{
		}

		Message(Genode::Heap& heap, const Socket_wrapper& socket)
			: _heap(heap)
		{
			Header header;
			Genode::memset(&header, 0, sizeof(header));
			socket.read(&header, sizeof(header));

			if (header.protocol_version != PROTOCOL_VERSION) {
				Genode::error("Bad HSM protocol version. Expected ",
				              static_cast<uint32_t>(PROTOCOL_VERSION),
				              ", actual ",
				              static_cast<uint32_t>(header.protocol_version));
				throw Bad_hsm_protcol_version_exception();
			}

			_command = static_cast<Command>(header.command_id);
			_status = header.status;
			_data_length = header.data_length;

			if (_data_length > 0) {
				if (!_heap.alloc(_data_length, reinterpret_cast<void**>(&_data_buffer))) {
					Genode::error("Cannot allocate message data buffer");
					throw Memory_alloc_exception();
				}

				socket.read(_data_buffer, _data_length);
			}

			uint8_t* current_buffer = _data_buffer;
			uint32_t current_length = _data_length;

			while (current_length > 0) {
				auto p = new (_heap) Payload(_heap, current_buffer, current_length);
				_payloads.enqueue(*p);
				current_buffer += p->length();
				current_length -= p->length();
			}
		}

		~Message()
		{
			_payloads.for_each([&] (Payload& p) {
				_payloads.remove(p);
				destroy(_heap, &p);
			});

			if (_data_buffer != nullptr) {
				_heap.free(_data_buffer, _data_length);
			}
		}

		Payload const *payload_by_type(const Data_type data_type) const
		{
			Payload* payload = nullptr;

			_payloads.for_each([&] (Payload& p) {
				if (p.data_type() == data_type) {
					payload = &p;
				}
			});
			return payload;
		}

		uint32_t data_length() const
		{
			uint32_t value = 0;
			_payloads.for_each([&] (Payload& p) {
				value += p.length();
			});
			return value;
		}

		uint32_t length() const
		{
			return data_length() + sizeof(Header);
		}

		void add(const Data_type data_type, const uint8_t* payload_buffer, const uint32_t payload_length)
		{
			auto payload = new (_heap) Payload(_heap, data_type, payload_buffer, payload_length);
			_payloads.enqueue(*payload);
		}

		void add(const Data_type data_type, const char* value)
		{
			add(data_type, reinterpret_cast<const uint8_t*>(value), strlen(value));
		}

		void add(const Data_type data_type, const uint32_t value)
		{
			// FIXME: don't we need to consider endianess here?!
			add(data_type, reinterpret_cast<const uint8_t*>(&value), sizeof(value));
		}

		void write_to_socket(const Socket_wrapper& socket) const
		{
			Header header;
			Genode::memset(&header, 0, sizeof(header));
			header.protocol_version = PROTOCOL_VERSION;
			header.command_id = _command;
			header.data_length = data_length();
			const size_t header_size { sizeof (header) };
			socket.write(&header, header_size);

			_payloads.for_each([&] (Payload& p) {
				p.write_to_socket(socket);
			});
		}
};

